"""
Daline体系中所有的单元类型定义,unit负责定义接口,frame负责框架和执行。所有业务单元均派生于以下基础单元之一
UnitBase: 单元基类,作为PU,TU等单元父类,不直接作为业务单元父类
EdgeUnit: 常驻业务服务的基类,派生常驻型后台服务
TaskUnit: 一次性或定期任务型脚本基类,做一次性或定期数据处理任务,不直接对外提供服务接口
TradeStrategyUnit: 量化交易策略单元基类,包含策略回测,实盘托管等接口定义
任何业务单元仅支持从预设单元类型中派生,单实例。按规范要求实现业务逻辑并在规定接口中返回数据结果,并要按规范要求编写docstring,代码编写及各类命名遵循编码规范,pylint无报错。
"""
import os
import json
import datetime
from logging import Logger
from abc import ABC, abstractmethod


class DalEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime) or \
                isinstance(obj, datetime.date) or \
                isinstance(obj, datetime.time):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class UnitBase(ABC):
    """
    所有单元基类,主要处理日志、接口路由等基础功能

    Attributes:
        _logger (logger): 日志实例
    """

    def __init__(self):
        """
        初始化
        """
        self._curcall_request = None
        self._logger = None
        self._logger_fresh_fun = None
        self._folder_temp = ''
        self._folder_nfs = ''
        self._request_func = None
        self._request_glb_func = None
        self._reg_request = {}
        self._reg_subscription = {}
        self._token = {}
        self._ws_udefine_emit_fun = None
        self._edge_info = {}
        self._hw_usage = {}

    @abstractmethod
    def test(self) -> bool:
        """
        各业务单元自定义单元测试部分

        Returns:
            True: 自动测试成功
            False: 自动测试失败
        """
        return True

    @abstractmethod
    def set_sys_notify(self, sys_notify):
        """
        系统对该单元的通知，心跳包返回，每分钟更新
        """
        return

    @classmethod
    def i18n_retinfo(cls, ret_info) -> str:
        """
        国际化返回提示信息,负责单元实例范围内的返回信息国际化,远期将各语言版本分别用对应文件加载
        """
        i18n = os.getenv('I18N', 'en')
        i18n_all = {
            'cn': {
                'success': "成功"
            }
        }
        return i18n_all.get(i18n, {}).get(ret_info, ret_info)

    @property
    def curcall_request(self) -> any:
        """
        获取当前调入的request实例
        """
        return self._curcall_request
        # return logging.getLogger()

    @curcall_request.setter
    def curcall_request(self, value):
        """
        设置当前调入的request实例
        """
        self._curcall_request = value

    @property
    def logger(self) -> Logger:
        """
        使用logger实例
        """
        return self._logger
        # return logging.getLogger()

    @logger.setter
    def logger(self, value):
        """
        设置本单元所用logger实例

        Parameters:
            logger (logger): logger实例
        """
        self._logger = value

    def set_logger_fresh_fun(self, fun):
        """
        设置模块日志刷新设置函数,引入会打印日志的第三方库时调用logger_fresh_set()接管

        Parameters:
            fun (fun): 日志刷新设置函数
        """
        self._logger_fresh_fun = fun

    def logger_fresh_set(self):
        """
        引入会打印日志的第三方库时调用,做日志打印接管

        Parameters:
            fun (fun): 日志刷新设置函数
        """
        self._logger_fresh_fun()

    @property
    def folder_temp(self) -> str:
        """
        获取临时存储目录
        """
        return self._folder_temp

    @folder_temp.setter
    def folder_temp(self, path) -> str:
        """
        设置本单元所用临时存储目录,仅做临时缓存使用,仅本单元可见,单元重启后数据可能丢失

        Parameters:
            path (string): 临时存储目录的绝对路径
        """
        self._folder_temp = path

    @property
    def folder_nfs(self) -> str:
        """
        获取单元用NFS存储目录
        """
        return self._folder_nfs

    @folder_nfs.setter
    def folder_nfs(self, path) -> str:
        """
        设置本单元本版本所用NFS存储目录,永久存储,同名业务单元的多个实例共享,单元重启后数据不丢失

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_nfs = path

    @property
    def folder_nfs_unit(self) -> str:
        """
        获取本单元NFS存储目录,本单元各版本存储目录都在本目录下
        """
        return os.path.split(os.path.split(self.folder_nfs)[0])[0]+"/"

    @property
    def folder_nfs_author(self) -> str:
        """
        获取本作者NFS存储目录,各单元存储目录都在本作者目录下
        """
        return os.path.split(os.path.split(self.folder_nfs_unit)[0])[0]+"/"

    def register_request(self, unit, version, fidx):
        """
        注册对其他单元接口的调用
        Parameters:
            unit (string): 其他单元名,如'pu_fakeif'
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example'
            method (string): 调用方式,如'get','post'等
        Returns:
            调用句柄
        """
        handle = len(self._reg_request)
        self._reg_request[handle] = {}
        self._reg_request[handle]['unit'] = unit
        self._reg_request[handle]['version'] = version
        self._reg_request[handle]['fidx'] = fidx
        return handle

    @property
    def registerd_request(self) -> dict:
        """
        所有对其他单元调用注册
        """
        return self._reg_request

    def set_request_func(self, func):
        """
        框架设置单元调用其他单元的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._request_func = func

    def call_unit(self, handle, index) -> tuple:
        """
        调用其他单元接口
        Parameters:
            func (method): 入口函数
        """
        if handle not in self._reg_request:
            raise Exception(f"Handle[{handle}] not registered!")
        return self._request_func(self._reg_request[handle], index)

    @property
    def token(self) -> dict:
        """
        开发者token,仅供特定场景使用,不要直接使用,调用其他单元通过registerd_request
        """
        return self._token

    def set_token(self, token):
        """
        设置开发者token
        """
        self._token = token

    @property
    def edgeinfo(self) -> dict:
        """
        边缘设备基础信息
        """
        return self._edge_info

    def set_edgeinfo(self, edgeinfo):
        """
        设置边缘设备基础信息
        """
        self._edge_info = edgeinfo

    @property
    def hwusage(self) -> dict:
        """
        边缘设备基础信息
        """
        return self._hw_usage

    def set_hwusage(self, hwusage):
        """
        更新设备使用率指标
        """
        self._hw_usage = hwusage

    def set_ws_udefine_emit_fun(self, fun):
        """
        框架设置单元通过websocket发送单元自定义数据的回调接口
        Parameters:
            fun (method): 入口函数
        """
        self._ws_udefine_emit_fun = fun

    def ws_udefine_emit(self, unit, version, data):
        """
        通过websocket发送单元自定义数据
        """
        return self._ws_udefine_emit_fun(unit, version, data)

    def wsclient_connected(self, _sid, _wstag):
        """
        对其他单元订阅连入
        """
        return

    def wsclient_disconnected(self, _sid, _wstag):
        """
        对其他单元订阅断联
        """
        return

class EdgeUnit(UnitBase):
    """
    常驻业务单元基类
    """

    def __init__(self):
        """
        初始化
        """
        super().__init__()
        self._task_update_progress_func = None
        self._get_sub_byds_func = None

    @abstractmethod
    def custom_init(self) -> bool:
        """
        各业务单元自定义初始化部分,初始化完成后单元进入正常运行状态,如果出错会自动尝试重新调用此函数再进行初始化直至成功

        Returns:
            True: 初始化完成
            False: 初始化中有错误,会自动重新调用
        """
        return True

    def run_once(self) -> int:
        """
        仅运行一次的入口函数，返回执行结果
        """
        return 0

    @classmethod
    def i18n_retinfo(cls, ret_info) -> str:
        """
        国际化返回提示信息,负责单元范围内的返回信息国际化,远期将各语言版本分别用对应文件加载
        """
        sup_ret_info = super().i18n_retinfo(ret_info)
        i18n = os.getenv('I18N', 'en')
        i18n_all = {
            'cn': {
            }
        }
        return i18n_all.get(i18n, {}).get(ret_info, sup_ret_info)

    def set_task_update_progress_func(self, func):
        """
        框架设置更新任务进度回调函数
        Parameters:
            func (method): 入口函数
        """
        self._task_update_progress_func = func

    def task_update_progress(self, progress):
        """
        更新任务进度,progress为float型,0~100
        """
        if self._task_update_progress_func:
            self._task_update_progress_func(progress)

    def set_get_sub_byds_func(self, func):
        """
        框架设置获取按ds获取订阅列表函数
        Parameters:
            func (method): 入口函数
        """
        self._get_sub_byds_func = func

    def get_sub_byds(self, ds_name) -> list:
        """
        返回对ds_name的订阅列表
        """
        if self._get_sub_byds_func:
            return self._get_sub_byds_func(ds_name)
        return []
